// Snaps building placement to cursor and previews it
function autociv_showBuildingPlacementTerrainSnap(mousePosX, mousePosY)
{
	placementSupport.position = Engine.GetTerrainAtScreenPoint(mousePosX, mousePosY);

	if (placementSupport.mode === "wall")
	{
		// Including only the on-screen towers in the next snap candidate list is sufficient here, since the user is
		// still selecting a starting point (which must necessarily be on-screen). (The update of the snap entities
		// itself happens in the call to updateBuildingPlacementPreview below).
		placementSupport.wallSnapEntitiesIncludeOffscreen = false;
	}
	else
	{
		// Cancel if not enough resources
		if (placementSupport.template && Engine.GuiInterfaceCall("GetNeededResources", { "cost": GetTemplateData(placementSupport.template).cost }))
		{
			placementSupport.Reset();
			inputState = INPUT_NORMAL;
			return true;
		}

		let snapData = Engine.GuiInterfaceCall("GetFoundationSnapData", {
			"template": placementSupport.template,
			"x": placementSupport.position.x,
			"z": placementSupport.position.z,
		});
		if (snapData)
		{
			placementSupport.angle = snapData.angle;
			placementSupport.position.x = snapData.x;
			placementSupport.position.z = snapData.z;
		}
	}

	// Includes an update of the snap entity candidates
	updateBuildingPlacementPreview();

	return;
}

/**
 * Select a building from the construction panel given the current unit(s) selected.
 * @param return - true if buildings exist and is selected to place.
 */
function autociv_placeBuildingByGenericName(genericName)
{
	// Hack: fast check
	if (Engine.GetGUIObjectByName("unitConstructionPanel").hidden)
		return;

	let translatedGenericName = translate(genericName);
	let index = g_SelectionPanels.Construction.getItems().findIndex(item =>
		GetTemplateData(item).name.generic == translatedGenericName);

	if (index == -1)
		return;

	let unitConstructionButton = Engine.GetGUIObjectByName(`unitConstructionButton[${index}]`);

	if (!unitConstructionButton ||
		unitConstructionButton.hidden ||
		!unitConstructionButton.enabled ||
		!unitConstructionButton.onPress)
		return;

	unitConstructionButton.onPress();
	autociv_showBuildingPlacementTerrainSnap(mouseX, mouseY);

	return true;
}

function autociv_clearSelectedProductionQueues()
{
	// Hack: fast check
	if (Engine.GetGUIObjectByName("unitQueuePanel").hidden)
		return;

	g_Selection.toList().map(GetEntityState).forEach(entity =>
	{
		if (!entity.production)
			return;

		for (let item of entity.production.queue)
			Engine.PostNetworkCommand({
				"type": "stop-production",
				"entity": entity.id,
				"id": item.id
			});
	});
	return true;
}

function autociv_SetCorpsesMax(value)
{
	Engine.GuiInterfaceCall("autociv_SetCorpsesMax", value);
}

function autociv_SetAutotrain(active, entities = g_Selection.toList())
{
	Engine.GuiInterfaceCall("autociv_SetAutotrain", {
		"active": active,
		"entities": entities
	});
}

var g_autociv_hotkeys = {
	"autociv.session.building.autotrain.enable": function (ev)
	{
		autociv_SetAutotrain(true);
		return true;
	},
	"autociv.session.building.autotrain.disable": function (ev)
	{
		autociv_SetAutotrain(false);
		return true;
	},
	"autociv.open.autociv_settings": function (ev)
	{
		autocivCL.Engine.PushGuiPage("page_autociv_settings.xml");
		return true;
	},
	"autociv.session.production.queue.clear": function (ev)
	{
		autociv_clearSelectedProductionQueues();
		return true;
	},
	"autociv.session.minimap.expand.toggle": function (ev)
	{
		autociv_minimapExpand.toggle();
		return true;
	}
}

var g_autociv_hotkeysPrefixes = {
	"autociv.session.building.place.": function (ev, hotkeyPrefix)
	{
		let buildingGenericName = ev.hotkey.split(hotkeyPrefix)[1].replace(/_/g, " ");
		autociv_placeBuildingByGenericName(buildingGenericName);
		return true;
	},
	"autociv.session.entity.select.": function (ev, hotkeyPrefix)
	{
		let entityGenericName = ev.hotkey.split(hotkeyPrefix)[1].replace(/_/g, " ");
		autociv_select.entityWithGenericName(entityGenericName);
		return true;
	},
	"autociv.session.entity.by.class.select.": function (ev, hotkeyPrefix)
	{
		let entityClassesExpression = ev.hotkey.split(hotkeyPrefix)[1];
		autociv_select.entityWithClassesExpression(entityClassesExpression, true);
		return true;
	},
	"autociv.session.formation.set.": function (ev, hotkeyPrefix)
	{
		let formationName = ev.hotkey.split(hotkeyPrefix)[1];
		autociv_formation.set(formationName);
		return true;
	},
	"autociv.session.stance.set.": function (ev, hotkeyPrefix)
	{
		let stanceName = ev.hotkey.split(hotkeyPrefix)[1];
		autociv_stance.set(stanceName);
		return true;
	}
};

autociv_patchApplyN("handleInputBeforeGui", function (target, that, args)
{
	let [ev] = args;
	autociv_APM.add(ev);
	return target.apply(that, args);
})

autociv_patchApplyN("handleInputAfterGui", function (target, that, args)
{
	let [ev] = args;
	if ("hotkey" in ev && ev.type == "hotkeydown")
	{
		// Special case hotkeys
		for (let prefix in g_autociv_hotkeysPrefixes)
			if (ev.hotkey.startsWith(prefix))
				return !!g_autociv_hotkeysPrefixes[prefix](ev, prefix);

		// Hotkey with normal behaviour
		if (ev.hotkey in g_autociv_hotkeys)
			return !!g_autociv_hotkeys[ev.hotkey](ev);
	}
	return target.apply(that, args);
})

unitFilters.autociv_isNotWounded = entity =>
{
	let entState = GetEntityState(entity);
	return entState &&
		hasClass(entState, "Unit") &&
		entState.maxHitpoints &&
		100 * entState.hitpoints > entState.maxHitpoints * Engine.ConfigDB_GetValue("user", "gui.session.woundedunithotkeythreshold");
};

autociv_patchApplyN("getPreferredEntities", function (target, that, args)
{
	let [ents] = args;
	if (Engine.HotkeyIsPressed("autociv.selection.nowoundedonly"))
		return ents.filter(unitFilters.autociv_isNotWounded);

	return target.apply(that, args);
})
